/* SPDX-License-Identifier: GPL-2.0-only */
/*
 * Copyright (c) 2023 Qualcomm Innovation Center, Inc. All rights reserved.
 */

#include "dsi_pll.h"
/* Register Offsets from PHY base address */
#define PHY_CMN_CLK_CFG0	0x010
#define PHY_CMN_CLK_CFG1	0x014
#define PHY_CMN_GLBL_CTRL       0x018
#define PHY_CMN_CTRL_0          0x01C
#define PHY_CMN_CTRL_1          0x020
#define PHY_CMN_PLL_CNTRL	0x048
#define PHY_CMN_LDO_CNTRL	0x04c

#define PLL_IE_TRIM              0x0400
#define PLL_IP_TRIM              0x0404

#define PLL_IPTAT_TRIM           0x0410

#define PLL_CLKBUFLR_EN          0x041c

#define PLL_SYSCLK_EN_RESET      0x0428
#define PLL_RESETSM_CNTRL        0x042c
#define PLL_RESETSM_CNTRL2       0x0430
#define PLL_RESETSM_CNTRL3       0x0434
#define PLL_RESETSM_CNTRL4       0x0438
#define PLL_RESETSM_CNTRL5       0x043c
#define PLL_KVCO_DIV_REF1        0x0440
#define PLL_KVCO_DIV_REF2        0x0444
#define PLL_KVCO_COUNT1          0x0448
#define PLL_KVCO_COUNT2          0x044c
#define PLL_VREF_CFG1            0x045c

#define PLL_KVCO_CODE            0x0458

#define PLL_VCO_DIV_REF1         0x046c
#define PLL_VCO_DIV_REF2         0x0470
#define PLL_VCO_COUNT1           0x0474
#define PLL_VCO_COUNT2           0x0478
#define PLL_PLLLOCK_CMP1         0x047c
#define PLL_PLLLOCK_CMP2         0x0480
#define PLL_PLLLOCK_CMP3         0x0484
#define PLL_PLLLOCK_CMP_EN       0x0488
#define PLL_PLL_VCO_TUNE         0x048C
#define PLL_DEC_START            0x0490
#define PLL_SSC_EN_CENTER        0x0494
#define PLL_SSC_ADJ_PER1         0x0498
#define PLL_SSC_ADJ_PER2         0x049c
#define PLL_SSC_PER1             0x04a0
#define PLL_SSC_PER2             0x04a4
#define PLL_SSC_STEP_SIZE1       0x04a8
#define PLL_SSC_STEP_SIZE2       0x04ac
#define PLL_DIV_FRAC_START1      0x04b4
#define PLL_DIV_FRAC_START2      0x04b8
#define PLL_DIV_FRAC_START3      0x04bc
#define PLL_TXCLK_EN             0x04c0
#define PLL_PLL_CRCTRL           0x04c4

#define PLL_RESET_SM_READY_STATUS 0x04cc

#define PLL_PLL_MISC1            0x04e8

#define PLL_CP_SET_CUR           0x04f0
#define PLL_PLL_ICPMSET          0x04f4
#define PLL_PLL_ICPCSET          0x04f8
#define PLL_PLL_ICP_SET          0x04fc
#define PLL_PLL_LPF1             0x0500
#define PLL_PLL_LPF2_POSTDIV     0x0504
#define PLL_PLL_BANDGAP  0x0508


/* Bit definition of SSC control registers */
#define SSC_CENTER              BIT(0)
#define SSC_EN                  BIT(1)
#define SSC_FREQ_UPDATE         BIT(2)
#define SSC_FREQ_UPDATE_MUX	BIT(3)
#define SSC_UPDATE_SSC          BIT(4)
#define SSC_UPDATE_SSC_MUX      BIT(5)
#define SSC_START               BIT(6)
#define SSC_START_MUX           BIT(7)

/* Dynamic Refresh Control Registers */
#define DSI_DYNAMIC_REFRESH_PLL_CTRL0		(0x014)
#define DSI_DYNAMIC_REFRESH_PLL_CTRL1		(0x018)
#define DSI_DYNAMIC_REFRESH_PLL_CTRL2		(0x01C)
#define DSI_DYNAMIC_REFRESH_PLL_CTRL3		(0x020)
#define DSI_DYNAMIC_REFRESH_PLL_CTRL4		(0x024)
#define DSI_DYNAMIC_REFRESH_PLL_CTRL5		(0x028)
#define DSI_DYNAMIC_REFRESH_PLL_CTRL6		(0x02C)
#define DSI_DYNAMIC_REFRESH_PLL_CTRL7		(0x030)
#define DSI_DYNAMIC_REFRESH_PLL_CTRL8		(0x034)
#define DSI_DYNAMIC_REFRESH_PLL_CTRL9		(0x038)
#define DSI_DYNAMIC_REFRESH_PLL_CTRL10		(0x03C)
#define DSI_DYNAMIC_REFRESH_PLL_CTRL11		(0x040)
#define DSI_DYNAMIC_REFRESH_PLL_CTRL12		(0x044)
#define DSI_DYNAMIC_REFRESH_PLL_CTRL13		(0x048)
#define DSI_DYNAMIC_REFRESH_PLL_CTRL14		(0x04C)
#define DSI_DYNAMIC_REFRESH_PLL_CTRL15		(0x050)
#define DSI_DYNAMIC_REFRESH_PLL_CTRL16		(0x054)
#define DSI_DYNAMIC_REFRESH_PLL_CTRL17		(0x058)
#define DSI_DYNAMIC_REFRESH_PLL_CTRL18		(0x05C)
#define DSI_DYNAMIC_REFRESH_PLL_CTRL19		(0x060)
#define DSI_DYNAMIC_REFRESH_PLL_CTRL20		(0x064)
#define DSI_DYNAMIC_REFRESH_PLL_CTRL21		(0x068)
#define DSI_DYNAMIC_REFRESH_PLL_CTRL22		(0x06C)
#define DSI_DYNAMIC_REFRESH_PLL_CTRL23		(0x070)
#define DSI_DYNAMIC_REFRESH_PLL_CTRL24		(0x074)
#define DSI_DYNAMIC_REFRESH_PLL_CTRL25		(0x078)
#define DSI_DYNAMIC_REFRESH_PLL_CTRL26		(0x07C)
#define DSI_DYNAMIC_REFRESH_PLL_CTRL27		(0x080)
#define DSI_DYNAMIC_REFRESH_PLL_CTRL28		(0x084)
#define DSI_DYNAMIC_REFRESH_PLL_CTRL29		(0x088)
#define DSI_DYNAMIC_REFRESH_PLL_CTRL30		(0x08C)
#define DSI_DYNAMIC_REFRESH_PLL_CTRL31		(0x090)
#define DSI_DYNAMIC_REFRESH_PLL_UPPER_ADDR	(0x094)
#define DSI_DYNAMIC_REFRESH_PLL_UPPER_ADDR2	(0x098)

#define DSI_PHY_TO_PLL_OFFSET	(0x500)

enum {
	DSI_PLL_0,
	DSI_PLL_1,
	DSI_PLL_MAX
};

struct dsi_pll_div_table pll_14nm_dphy_lb[] = {
	{27270000, 30000000, 2, 11},
	{30000000, 33330000, 4, 5},
	{33330000, 37500000, 2, 9},
	{37500000, 40000000, 8, 2},
	{40000000, 42860000, 1, 15},
	{42860000, 46150000, 2, 7},
	{46150000, 50000000, 1, 13},
	{50000000, 54550000, 4, 3},
	{54550000, 60000000, 1, 11},
	{60000000, 66670000, 2, 5},
	{66670000, 75000000, 1, 9},
	{75000000, 85710000, 8, 1},
	{85710000, 100000000, 1, 7},
	{100000000, 120000000, 2, 3},
	{120000000, 150000000, 1, 5},
	{150000000, 200000000, 4, 1},
	{200000000, 300000000, 1, 3},
	{300000000, 600000000, 2, 1},
	{600000000, 1500000000, 1, 2}
};

struct dsi_pll_div_table pll_14nm_dphy_hb[] = {
	{68180000, 75000000, 2, 11},
	{75000000, 83330000, 4, 5},
	{83330000, 93750000, 2, 9},
	{93750000, 100000000, 8, 2},
	{100000000, 107140000, 1, 15},
	{107140000, 115380000, 2, 7},
	{115380000, 125000000, 1, 13},
	{125000000, 136360000, 4, 3},
	{136360000, 150000000, 1, 11},
	{150000000, 166670000, 2, 5},
	{166670000, 187500000, 1, 9},
	{187500000, 214290000, 8, 1},
	{214290000, 250000000, 1, 7},
	{250000000, 300000000, 2, 3},
	{300000000, 375000000, 1, 5},
	{375000000, 500000000, 4, 1},
	{500000000, 750000000, 1, 3},
	{750000000, 1500000000, 2, 2},
	{1500000000, 5000000000, 1, 1}
};

struct dsi_pll_div_table pll_14nm_cphy_lb[] = {
	{30000000, 37500000, 4, 5},
	{37500000, 50000000, 8, 2},
	{50000000, 60000000, 4, 3},
	{60000000, 75000000, 2, 5},
	{75000000, 100000000, 8, 1},
	{100000000, 120000000, 2, 3},
	{120000000, 150000000, 1, 5},
	{150000000, 200000000, 4, 1},
	{200000000, 300000000, 1, 3},
	{300000000, 600000000, 2, 1},
	{600000000, 1500000000, 1, 1}
};

struct dsi_pll_div_table pll_14nm_cphy_hb[] = {
	{75000000, 93750000, 4, 5},
	{93750000, 12500000, 8, 2},
	{125000000, 150000000, 4, 3},
	{150000000, 187500000, 2, 5},
	{187500000, 250000000, 8, 1},
	{250000000, 300000000, 2, 3},
	{300000000, 375000000, 1, 5},
	{375000000, 500000000, 4, 1},
	{500000000, 750000000, 1, 3},
	{750000000, 1500000000, 2, 1},
	{1500000000, 5000000000, 1, 1}
};
